import inspect
import logging

logger = logging.getLogger(__name__)


class EndpointHandler:
    def __init__(self, arg_definitions, arg_values, function):
        self.arg_definitions = arg_definitions
        self.arg_values = arg_values
        self.function = function
        self.values = {}
        self.fields = arg_definitions.get('fields')

    def evaluate_args(self):
        for field in self.fields:
            field.value = self.arg_values.get(field.name)
            field.handler = self
            field.evaluate()

    def get_fields(self):
        self.evaluate_args()
        body, params, fields = {}, {}, {}
        for x in [field for field in self.fields if field.body and field.value]:
            body.update(x.to_dict())
        for x in [field for field in self.fields if field.param and field.value]:
            params.update(x.to_dict())
        for x in [field for field in self.fields]:
            fields.update(x.to_dict())
        return Fields(
            body=body,
            params=params,
            fields=fields
        )


class Fields:
    def __init__(self, body: dict = None, params: dict = None, fields: dict = None):
        self.body = body
        self.params = params
        if fields:
            for k, v in fields.items():
                self.__setattr__(k, v)


class Field:
    def __init__(self, name, required=True, condition=None, param=False, body=True):
        self.name = name
        self.required = required
        self.condition = condition
        self.param = param
        self.body = body
        self.value = None
        self.handler = EndpointHandler({}, {}, '')

    def evaluate(self):
        # Always required field without value
        if self.required and not self.condition and not self.value:
            raise KeyError(f'{self.name} is required for {self.handler.function}')
        # Conditional required field where condition is met but with no value
        elif self.required and self.condition and self.check_condition() and not self.value:
            raise KeyError(f'{self.name} is required for {self.handler.function} when >> {self.condition} << is True')
        # Conditional field where the condition is not met but has a value
        elif self.condition and not self.check_condition() and self.value:
            raise KeyError(
                f'{self.name} is not allowed for {self.handler.function} when >> {self.condition} << is False')

    def check_condition(self):
        for field in self.handler.fields:
            locals().update(field.to_dict())
        locals().update(self.handler.arg_values)
        if eval(self.condition):
            return True
        return False

    def to_dict(self):
        return {self.name: self.value}


fields = Fields()


def endpoint(*, fields: list = None, pre_hook=None, post_hook=None):
    endpoint_arg_definitions = {k: v for k, v in inspect.getargvalues(inspect.currentframe()).locals.items() if
                                '_' != k[0]}

    def decorator(func):
        def decorated(_self, *args, **kwargs):
            # Get argument values passed to the function
            arg_values = {k: v for k, v in inspect.getargvalues(inspect.currentframe()).locals.items() if '_' != k[0]}
            # Update kwargs values with any values on the function's parent object
            arg_values['kwargs'].update({k: v for k, v in _self.__dict__.items() if '_' != k[0]})
            # Set the handler to check fields
            handler = EndpointHandler(arg_definitions=endpoint_arg_definitions,
                                      arg_values=arg_values['kwargs'],
                                      function=f'{_self.__class__.__name__}.{func.__name__}')
            # Update function globals so fields are available in the function
            original_globals = func.__globals__.copy()
            try:
                _fields = handler.get_fields()
                if pre_hook and callable(pre_hook):
                    pre_hook(_fields)
                func.__globals__.update({'fields': _fields})
                func_response = func(_self)
                if post_hook and callable(post_hook):
                    post_hook(func_response)
            finally:
                func.__globals__.update(original_globals)
            return func_response

        return decorated

    return decorator
    # Verify all required fields are present either on object or in arguments
    # Match all fields to values
    # Add values to Fields object
    # for field in [y for x in [required_fields, optional_fields] for y in x]:
    # pass



# -*- coding: utf-8 -*-
import logging
from .fabled_rest import FabledREST
from .handlers.endpoints import endpoint, fields, Field

"""Top-level package for Fabled REST."""

__author__ = """Zachary Young"""
__email__ = 'solve@fabled.dev'
__version__ = '0.2.1'

FabledREST = FabledREST
endpoint = endpoint
fields = fields
logging.getLogger('fabled_REST').addHandler(logging.NullHandler())



import logging

logger = logging.getLogger(__name__)


class ApiError(Exception):
    pass


class ErrorHandler:
    def __init__(self, response):
        self.response = response
        self._check_code()

    def _check_code(self):
        sc = self.response.status_code
        try:
            if sc < 300:
                return None
            else:
                raise ApiError
        except ApiError:
            logger.warning(f'---- API Error ----\n'
                           f'Status Code {sc}: {self.response.text}\n'
                           f'URL: {self.response.request.method} {self.response.request.url}\n'
                           f'Data: {self.response.request.body}\n'
                           f'Headers: {self.response.request.headers}\n'
                           f'--------------------')



import copy
import json
import logging
from tqdm import tqdm
from .local.progress.spinner import PixelSpinner2
import requests
try:
    from fabled_rest.handlers.errors import ErrorHandler
except ImportError:
    from fabled_rest.fabled_rest.handlers.errors import ErrorHandler


class ApiClient:
    """Chainable, magical class helps you make requests to RESTful services"""

    HTTP_METHODS = ['get', 'options', 'head', 'post', 'put', 'patch', 'delete']

    def __init__(self, name=None, parent=None, append_slash=False, **kwargs):
        """Constructor

        Arguments:
            name -- name of node
            parent -- parent node for chaining
            append_slash -- flag if you want a trailing slash in urls
            **kwargs -- `requests` session be initiated with if any available
        """
        self._name = name
        self._parent = parent
        self._append_slash = append_slash
        self._session = requests.session()
        for k, v in kwargs.items():
            orig = getattr(self._session, k, None)
            if isinstance(orig, dict):
                orig.update(v)
            elif orig:
                setattr(self._session, k, v)

    def _spawn(self, name):
        """Returns a shallow copy of current `ApiClient` instance as nested child

        Arguments:
            name -- name of child
        """
        child = copy.copy(self)
        child._name = name
        child._parent = self
        return child

    def __getattr__(self, name):
        """Here comes some magic. Any absent attribute typed within class
        falls here and return a new child `ApiClient` instance in the chain.
        """
        # Ignore specials (Otherwise shallow copying causes infinite loops)
        if name.startswith('__'):
            raise AttributeError(name)
        return self._spawn(name)

    def __iter__(self):
        """Iterator implementation which iterates over `ApiClient` chain."""
        current = self
        while current:
            if current._name:
                yield current
            current = current._parent

    def _chain(self, *args):
        """This method converts args into chained ApiClient instances

        Arguments:
            *args -- array of string representable objects
        """
        chain = self
        for arg in args:
            chain = chain._spawn(str(arg))
        return chain

    def _close_session(self):
        """Closes session if exists"""
        if self._session:
            self._session.close()

    def __call__(self, *args):
        """Here comes second magic. If any `ApiClient` instance called it
        returns a new child `Hammock` instance in the chain
        """
        return self._chain(*args)

    def _url(self, *args):
        """Converts current `ApiClient` chain into a url string

        Arguments:
            *args -- extra url path components to tail
        """
        path_comps = [mock._name for mock in self._chain(*args)]
        url = "/".join(reversed(path_comps))
        if self._append_slash:
            url = url + "/"
        return url

    def __repr__(self):
        """ String representation of current `ApiClient` chain"""
        return self._url()

    def _request(self, method, *args, **kwargs):
        """
        Makes the HTTP request using requests module
        """
        return self._session.request(method, self._url(*args), **kwargs)


def bind_method(method):
    """Bind `requests` module HTTP verbs to `ApiClient` class as
    static methods."""

    def aux(hammock, *args, post_hook=None, **kwargs):
        response = hammock._request(method, *args, **kwargs)
        if post_hook and callable(post_hook):
            response = post_hook(response)
        ErrorHandler(response)
        return response

    return aux


for m in ApiClient.HTTP_METHODS:  # todo include post_hook in bind method
    setattr(ApiClient, m.upper(), bind_method(m))


class FabledREST(object):
    def __init__(self, base_url, client_data: dict = None, **kwargs):
        if base_url:
            if not client_data: client_data = {}
            self.client = ApiClient(base_url, **client_data)
        if kwargs:
            self.update_self(**kwargs)
        self._name = self.__class__.__name__

    def __str__(self):
        return f"({self._name}) {str(self.dict())}"

    def __dir__(self):
        return self.__dict__.keys()

    def __call__(self, *args, **kwargs):
        clone = copy.copy(self)
        if kwargs:
            return clone.update_self(**kwargs)
        return clone

    def __getattribute____(self, item):
        clone = copy.copy(self)
        return object.__getattribute__(clone, item)

    def json(self, include_only=None):
        return json.dumps(self.dict(include_only=include_only))

    def dict(self, include_only: list = None):
        if len(self.__dict__) < 1:
            return {}
        if include_only:
            include = {k: v for k, v in self.__dict__.items() if k in include_only}
            return {k: v for k, v in include.items() if v and k[0] != '_' and k != 'client'}
        return {k: v for k, v in self.__dict__.items() if v and k[0] != '_' and k != 'client'}

    def update_self(self, **kwargs):
        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)
        return self

    @staticmethod
    def progress(*args, **kwargs):
        return tqdm(*args, **kwargs)

    @staticmethod
    def spinner(*args, **kwargs):
        return PixelSpinner2(*args, **kwargs)


    @property
    def logger(self):
        import sys
        name = '.'.join([
            self.__module__,
            self.__class__.__name__
        ])
        logging.basicConfig(
            level=logging.DEBUG,
            format="[%(asctime)s] %(levelname)s %(name)s:%(lineno)d %(message)s",
            datefmt="%H:%M:%S",
            stream=sys.stdout)
        return logging.getLogger(name)



# Copyright (c) 2012 Giorgos Verigakis <verigak@gmail.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import division, print_function

from collections import deque
from datetime import timedelta
from math import ceil
from sys import stderr
try:
    from time import monotonic
except ImportError:
    from time import time as monotonic



__version__ = '1.5'

HIDE_CURSOR = '\x1b[?25l'
SHOW_CURSOR = '\x1b[?25h'


class Infinite(object):
    file = stderr
    sma_window = 10         # Simple Moving Average window
    check_tty = False
    hide_cursor = True

    def __init__(self, message='', **kwargs):
        self.index = 0
        self.start_ts = monotonic()
        self.avg = 0
        self._avg_update_ts = self.start_ts
        self._ts = self.start_ts
        self._xput = deque(maxlen=self.sma_window)
        for key, val in kwargs.items():
            setattr(self, key, val)

        self._width = 0
        self.message = message

        if self.file and self.is_tty():
            if self.hide_cursor:
                print(HIDE_CURSOR, end='', file=self.file)
            print(self.message, end='', file=self.file)
            self.file.flush()

    def __getitem__(self, key):
        if key.startswith('_'):
            return None
        return getattr(self, key, None)

    @property
    def elapsed(self):
        return int(monotonic() - self.start_ts)

    @property
    def elapsed_td(self):
        return timedelta(seconds=self.elapsed)

    def update_avg(self, n, dt):
        if n > 0:
            xput_len = len(self._xput)
            self._xput.append(dt / n)
            now = monotonic()
            # update when we're still filling _xput, then after every second
            if (xput_len < self.sma_window or
                    now - self._avg_update_ts > 1):
                self.avg = sum(self._xput) / len(self._xput)
                self._avg_update_ts = now

    def update(self):
        pass

    def start(self):
        pass

    def clearln(self):
        if self.file and self.is_tty():
            print('\r\x1b[K', end='', file=self.file)

    def write(self, s):
        if self.file and self.is_tty():
            line = self.message + s.ljust(self._width)
            print('\r' + line, end='', file=self.file)
            self._width = max(self._width, len(s))
            self.file.flush()

    def writeln(self, line):
        if self.file and self.is_tty():
            self.clearln()
            print(line, end='', file=self.file)
            self.file.flush()

    def finish(self):
        if self.file and self.is_tty():
            self.write('Complete')
            print(file=self.file)
            if self.hide_cursor:
                print(SHOW_CURSOR, end='', file=self.file)

    def is_tty(self):
        return self.file.isatty() if self.check_tty else True

    def next(self, n=1):
        now = monotonic()
        dt = now - self._ts
        self.update_avg(n, dt)
        self._ts = now
        self.index = self.index + n
        self.update()

    def iter(self, it):
        with self:
            for x in it:
                yield x
                self.next()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()


class Progress(Infinite):
    def __init__(self, *args, **kwargs):
        super(Progress, self).__init__(*args, **kwargs)
        self.max = kwargs.get('max', 100)

    @property
    def eta(self):
        return int(ceil(self.avg * self.remaining))

    @property
    def eta_td(self):
        return timedelta(seconds=self.eta)

    @property
    def percent(self):
        return self.progress * 100

    @property
    def progress(self):
        return min(1, self.index / self.max)

    @property
    def remaining(self):
        return max(self.max - self.index, 0)

    def start(self):
        self.update()

    def goto(self, index):
        incr = index - self.index
        self.next(incr)

    def iter(self, it):
        try:
            self.max = len(it)
        except TypeError:
            pass

        with self:
            for x in it:
                yield x
                self.next()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click>=7.0', 'requests>=2.22.0', 'tqdm==4.36.1']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Zachary Young",
    author_email='solve@fabled.dev',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
    ],
    description="Package designed to make creating API SDKs easier.",
    entry_points={
        'console_scripts': [
            'fabled_rest=fabled_rest.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='fabled_rest',
    name='fabled_rest',
    packages=find_packages(include=['fabled_rest', 'fabled_rest.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/zpyoung/fabled_rest',
    version='0.2.1',
    zip_safe=False,
)
